<?php
/**
 * @package   AdminTools
 * @copyright 2010-2021 Akeeba Ltd / Nicholas K. Dionysopoulos
 * @license   GNU General Public License version 3, or later
 */

namespace Akeeba\Component\AdminTools\Administrator\Model;

defined('_JEXEC') or die;

use Akeeba\Component\AdminTools\Administrator\Controller\Mixin\TempsuperuserChecks;
use Akeeba\Component\AdminTools\Administrator\Table\Mixin\NoSuperUsersCheckFlags;
use Exception;
use Joomla\CMS\Access\Access;
use Joomla\CMS\Application\CMSApplication;
use Joomla\CMS\Date\Date;
use Joomla\CMS\Factory;
use Joomla\CMS\Form\FormFactoryInterface;
use Joomla\CMS\Language\Text;
use Joomla\CMS\MVC\Factory\MVCFactoryInterface;
use Joomla\CMS\MVC\Model\AdminModel;
use Joomla\CMS\User\User;
use Joomla\CMS\User\UserFactoryInterface;
use Joomla\CMS\User\UserHelper;
use RuntimeException;

class TempsuperuserModel extends AdminModel
{
	use NoSuperUsersCheckFlags;
	use TempsuperuserChecks;

	/**
	 * Cache of user group IDs with Super User privileges
	 *
	 * @var   array
	 * @since 5.3.0
	 */
	protected $superUserGroups = [];

	public function __construct($config = [], MVCFactoryInterface $factory = null, FormFactoryInterface $formFactory = null)
	{
		parent::__construct($config, $factory, $formFactory);

		$this->_parent_table = '';
	}

	/**
	 * @inheritDoc
	 */
	public function getForm($data = [], $loadData = true)
	{
		$pk       = (int) $this->getState($this->getName() . '.id');
		$id       = $data['user_id'] ?? $pk;
		$user     = empty($id) ? new User() : Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($id);
		$formName = (!empty($id) && ($user->id == $id)) ? 'tempsuperuser' : 'tempsuperuser_wizard';

		$form = $this->loadForm(
			'com_admintools.' . $formName,
			$formName,
			[
				'control'   => 'jform',
				'load_data' => $loadData,
			]
		) ?: false;

		if (empty($form))
		{
			return false;
		}

		return $form;
	}

	public function save($data)
	{
		try
		{
			$this->setNoCheckFlags(true);

			$isNew = empty($data['user_id'] ?? null);

			$data['user_id'] = ($data['user_id'] ?? null) ?: $this->getUserFromData($data);
		}
		catch (Exception $e)
		{
			$this->setError($e->getMessage());

			return false;
		}
		finally
		{
			$this->setNoCheckFlags(false);
		}

		try
		{
			$table   = $this->getTable();
			$context = $this->option . '.' . $this->name;
			$app     = Factory::getApplication();

			$key = $table->getKeyName();
			$pk  = (isset($data[$key])) ? $data[$key] : (int) $this->getState($this->getName() . '.id');

			if (!$isNew)
			{
				$table->load($pk);
			}

			// Bind the data.
			if (!$table->bind($data))
			{
				$this->setError($table->getError());

				return false;
			}

			// Prepare the row for saving
			$this->prepareTable($table);

			// Check the data.
			if (!$table->check())
			{
				$this->setError($table->getError());

				return false;
			}

			// Trigger the before save event.
			$result = $app->triggerEvent($this->event_before_save, [$context, $table, $isNew, $data]);

			if (\in_array(false, $result, true))
			{
				$this->setError($table->getError());

				return false;
			}

			// Store the data.
			try
			{
				if ($isNew)
				{
					$this->getDbo()->insertObject($table->getTableName(), $table, $table->getKeyName());
				}
				else
				{
					$this->getDbo()->updateObject($table->getTableName(), $table, $table->getKeyName(), true);
				}
			}
			catch (Exception $e)
			{
				$this->setError($e->getMessage());

				return false;
			}

			if (!$table->store())
			{
				$this->setError($table->getError());

				return false;
			}

			// Clean the cache.
			$this->cleanCache();

			// Trigger the after save event.
			$app->triggerEvent($this->event_after_save, [$context, $table, $isNew, $data]);
		}
		catch (Exception $e)
		{
			$this->setError($e->getMessage());

			return false;
		}

		if (isset($table->$key))
		{
			$this->setState($this->getName() . '.id', $table->$key);
		}

		$this->setState($this->getName() . '.new', $isNew);

		return true;
	}

	protected function canDelete($record)
	{
		$this->assertNotMyself($record->id);

		return parent::canDelete($record);
	}

	/**
	 * Loads the form data.
	 *
	 * This method has three modes of operation:
	 *
	 * - If there is saved form data in the session and the user_id (PK) matches we'll use that.
	 * - If we are editing an existing record we load the record.
	 * - Otherwise it's the wizard layout and we pre-fill it with some sane defaults.
	 *
	 * @return array|bool|\Joomla\CMS\Object\CMSObject|mixed
	 * @throws Exception
	 */
	protected function loadFormData()
	{
		/** @var CMSApplication $app */
		$app  = Factory::getApplication();
		$data = $app->getUserState('com_admintools.edit.tempsuperuser.data', []);
		$pk   = (int) $this->getState($this->getName() . '.id');
		$item = ($pk ? $this->getItem() : false) ?: [];

		// There's data saved in the session and the user_id in it matches what we're editing
		if (!empty($data) && ($item->user_id ?? null) == ($data['user_id'] ?? null))
		{
			$this->preprocessData('com_admintools.tempsuperuser', $data);

			return $data;
		}

		// First, let's try loading an existing item.
		$data = $item;
		$pk   = (int) $this->getState($this->getName() . '.id');

		if ($pk > 0)
		{
			$this->preprocessData('com_admintools.tempsuperuser', $data);

			return $data;
		}

		// No existing item. I am in the Wizard view. Let's preload it with some randomized, default values.
		$jDate           = new Date();
		$interval        = new \DateInterval('P15D');
		$password        = UserHelper::genRandomPassword(32);
		$superUserGroups = $this->getSuperUserGroups() ?: [8];

		return [
			'expiration' => $jDate->add($interval)->toRFC822(),
			'username'   => 'temp' . UserHelper::genRandomPassword(12),
			'password'   => $password,
			'password2'  => $password,
			'email'      => UserHelper::genRandomPassword(12) . '@example.com',
			'name'       => Text::_('COM_ADMINTOOLS_TEMPSUPERUSERS_LBL_DEFAULTNAME'),
			'groups'     => array_shift($superUserGroups),
		];
	}

	/**
	 * Returns all Joomla! user groups
	 *
	 * @return  array
	 *
	 * @since   5.3.0
	 */
	private function getSuperUserGroups()
	{
		if (!empty($this->superUserGroups))
		{
			return $this->superUserGroups;
		}

		// Get all groups
		$db    = $this->getDbo();
		$query = $db->getQuery(true)
			->select([$db->qn('id')])
			->from($db->qn('#__usergroups'));

		$this->superUserGroups = $db->setQuery($query)->loadColumn(0);

		// This should never happen (unless your site is very dead, in which case I feel terribly sorry for you...)
		if (empty($this->superUserGroups))
		{
			$this->superUserGroups = [];
		}

		$this->superUserGroups = array_filter($this->superUserGroups, function ($group) {
			return Access::checkGroup($group, 'core.admin');
		});

		return $this->superUserGroups;
	}

	private function getUserFromData($info)
	{
		$info['block']         = 0;
		$info['sendEmail']     = 0;
		$info['lastvisitDate'] = (new Date())->toSql();
		$info['activation']    = '';
		$info['otpKey']        = '';
		$info['otep']          = '';
		$info['requireReset']  = 0;

		$userId = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserByUsername($info['username'])->id;

		if (empty($userId))
		{
			return $this->createNewUser($info);
		}

		// Make sure I am not trying to edit myself
		if ($userId == Factory::getApplication()->getIdentity()->id)
		{
			throw new RuntimeException(Text::_('COM_ADMINTOOLS_TEMPSUPERUSERS_ERR_CANTEDITSELF'), 403);
		}

		// Get the existing user
		$user = Factory::getContainer()->get(UserFactoryInterface::class)->loadUserById($userId);

		// Make sure the user is a Super User
		if (!$user->authorise('core.admin'))
		{
			throw new RuntimeException(Text::_('COM_ADMINTOOLS_TEMPSUPERUSERS_ERR_NOTSUPER'), 500);
		}

		// Make sure the user was already blocked
		if (!$user->block)
		{
			throw new RuntimeException(Text::_('COM_ADMINTOOLS_TEMPSUPERUSERS_ERR_NOTBLOCKED'), 500);
		}

		// Apply changes to the existing user
		$user->bind($info);

		$saved = $user->save();

		if (!$saved)
		{
			throw new RuntimeException($user->getError());
		}

		return $userId;
	}

	private function createNewUser(&$info)
	{
		// Make sure $info['groups'] is defined and defines at least one Super User group
		$superUserGroups = $this->getSuperUserGroups();
		$usedSUGroups    = array_intersect($info['groups'], $superUserGroups);

		if (empty($usedSUGroups))
		{
			$this->setNoCheckFlags(false);

			throw new RuntimeException(Text::_('COM_ADMINTOOLS_TEMPSUPERUSERS_ERR_NOTASUPERUSER'), 500);
		}

		// Create a new user
		$user = new User();

		// Set the user's default language to whatever the site's current language is
		$info['params'] = [
			'language' => Factory::getApplication()->get('language'),
		];

		$user->bind($info);

		$saved = $user->save();

		if (!$saved)
		{
			$this->setNoCheckFlags(false);

			throw new RuntimeException($user->getError());
		}

		$this->addUserToSafeId($user->id);

		return $user->id;
	}

	/**
	 * Adds a new user into the list of "safe ids", otherwise at the next session load it will be disabled by the
	 * feature "Monitor Super User accounts"
	 *
	 * @param   int  $userid  ID of the new user that should be injected into the list
	 */
	private function addUserToSafeId($userid)
	{
		$db    = $this->getDbo();
		$query = $db->getQuery(true)
			->select($db->quoteName('value'))
			->from($db->quoteName('#__admintools_storage'))
			->where($db->quoteName('key') . ' = ' . $db->quote('superuserslist'));
		$db->setQuery($query);

		try
		{
			$jsonData = $db->loadResult();
		}
		catch (Exception $e)
		{
			return;
		}

		$userList = [];

		if (!empty($jsonData))
		{
			$userList = json_decode($jsonData, true);
		}

		$userList[] = $userid;

		$data = json_encode($userList);

		$query = $db->getQuery(true)
			->delete($db->quoteName('#__admintools_storage'))
			->where($db->quoteName('key') . ' = ' . $db->quote('superuserslist'));
		$db->setQuery($query);
		$db->execute();

		$object = (object) [
			'key'   => 'superuserslist',
			'value' => $data,
		];

		$db->insertObject('#__admintools_storage', $object);
	}

}